from rest_framework import serializers
from rest_framework.validators import UniqueTogetherValidator
from django.contrib.auth import get_user_model
from .core import get_global_history_model


class GenericSerializer(serializers.ModelSerializer):
    class Meta:
        model = None
        fields = "__all__"
        depth = 1


class GenericHistorySerializer(serializers.ModelSerializer):
    action = serializers.SerializerMethodField()
    user_link = serializers.SerializerMethodField()
    date = serializers.SerializerMethodField()
    record_link = serializers.SerializerMethodField()
    model_list_link = serializers.SerializerMethodField()
    record = serializers.SerializerMethodField()

    def get_action(self, global_history):
        if global_history.action == "ED":
            return f"""
            <a class="modal-open get-object-changes" href="#" data-url="{global_history.get_absolute_url()}"
            data-modal="#modal-changes"><i class="fas fa-pen hist-edit"></i></a>
            """
        elif global_history.action == "NE":
            return '<i class="fas fa-plus hist-add"></i>'
        elif global_history.action == "DE":
            return '<i class="fas fa-trash hist-del"></i>'

    def get_user_link(self, global_history):
        return global_history.user.to_datatable()

    def get_date(self, global_history):
        return global_history.date.strftime("%Y-%m-%d %H:%M:%S")

    def get_record(self, global_history):
        return global_history.content_object.history_object.name

    def get_record_link(self, global_history):
        if not global_history.action == "DE":
            return global_history.content_object.history_object.to_datatable()
        else:
            return global_history.record_name

    def get_model_list_link(self, global_history):
        return global_history.content_object.history_object.__class__.href_to_datalist()

    class Meta:
        model = get_global_history_model()
        fields = [
            "action",
            "user_link",
            "date",
            "record_link",
            "model_list_link",
            "record",
        ]
        depth = 1


class GenericUserSerializer(serializers.ModelSerializer):
    def create(self, validated_data):
        user = get_user_model().objects.create_user(**validated_data)
        return user

    class Meta:
        model = get_user_model()
        fields = (
            "username",
            "first_name",
            "last_name",
            "email",
            "password",
            "style",
        )

        validators = [
            UniqueTogetherValidator(
                queryset=get_user_model().objects.all(), fields=["username", "email"]
            )
        ]



from django.db import models
from django.conf import settings
from django.core.cache import cache
from django.urls import reverse_lazy, reverse
from django.utils.translation import gettext_lazy as _
from django.contrib.auth.models import AbstractUser

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType

from simple_history.models import HistoricalRecords

from .core import (
    get_global_history_model,
    get_model,
    get_history_change_model,
    str_list_to_str,
)


class GenericStyle(models.Model):
    """Classe abstraite générique représentant un fichier de style CSS"""

    name = models.CharField(_("Style"), max_length=55)
    static_path = models.CharField(_("Chemin statique du style"), max_length=255)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name


class GenericUser(AbstractUser):
    """Classe abstraite générique représentant un utilisateur"""

    # style = models.ForeignKey(
    #     settings.STYLE_MODEL, on_delete=models.CASCADE, null=True, blank=True,
    #     verbose_name=_("Style"))

    class Meta:
        abstract = True

    def get_absolute_url(self):
        """Retourne un lien vers le profile de l'utilisateur"""
        return reverse("profile", args=[str(self.id)])

    def to_datatable(self):
        """Retourne la vue pour l'affichage de l'utilisateur dans
        les 'datatable'
        """
        return f'<a href="{self.get_absolute_url()}">{str(self)}</a>'


class GenericModel(models.Model):
    """Classe abstraite générique dont peuvent hériter des modèles pour
    bénéficier de la génération automatique des vues django_bkendoz
    """

    # LINK_FIELD = "name"
    name = models.CharField(_("Nom"), max_length=255, default="")
    views_struct = {}

    class Meta:
        abstract = True
        verbose_name = "GenericModel"

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        cache.delete("omnibar:" + self.__class__.__name__)
        cache.delete_pattern(self.__class__.__name__ + ":viewset:*:*")
        return super().save(*args, **kwargs)

    def get_absolute_url(self):
        """Renvoie un lien vers une vue dans l'ordre suivant :
        * detail
        * datatable
        * list
        * sinon str vide

        Returns:
            str: href
        """
        app = self.__class__._meta.app_label.lower()
        model = str(self.__class__.__name__).lower()

        if "detail" in self.__class__.get_views_struct():
            return reverse(f"{app}:{model}-detail", args=[str(self.id)])

        if "datatable" in self.__class__.get_views_struct():
            return reverse(f"{app}:{model}-datatable")

        if "list" in self.__class__.get_views_struct():
            return reverse(f"{app}:{model}-list")

        return ""

    def to_datatable(self):
        """Retourne la vue pour l'affichage de l'instance du modèle
        dans les 'datatable'"""

        return f'<a href="{self.get_absolute_url()}">{str(self)}</a>'

    @classmethod
    def href_to_datalist(cls):
        """Retourne un lien vers l'affichage en liste du modèle"""

        verbose_name = cls._meta.verbose_name
        if "datatable" in cls.get_views_struct():
            path_name = cls.get_url("datatable")
        elif "list" in cls.get_views_struct():
            path_name = cls.get_url("list")
        else:
            return verbose_name

        return f'<a href="{reverse(path_name)}">{verbose_name}</a>'

    @classmethod
    def to_omnibar_dict(cls, qs=None):
        """Export des objets d'un modèle pour l'omnibar avec
        type, handler, param0, param1
        """

        qs = qs or cls.objects.all()
        items = []
        for obj in qs:
            items.append(
                {
                    "type": cls._meta.verbose_name,
                    "handler": "toDetail",
                    "param0": obj.get_omnibar_ref(),
                    "param1": obj.get_absolute_url(),
                }
            )
        return items

    def get_omnibar_ref(self):
        """Retourne une référence pour l'affichage dans l'omnibar, dans
        l'ordre de priorité :
        * omnibar
        * référence
        * name
        * str(self)
        """

        cls = self.__class__
        omnibar_ref = cls.get_view_data("omnibar")
        if omnibar_ref and hasattr(self, omnibar_ref):
            return getattr(self, omnibar_ref)
        elif hasattr(self, "reference"):
            return getattr(self, "reference")
        elif hasattr(self, "name"):
            return self.name
        else:
            return str(self)

    @classmethod
    def get_views_struct(cls):
        """Retourne la structure de vue du modèle."""

        return cls.views_struct

    @classmethod
    def get_view_data(cls, view, prop="fields"):
        """Retourne les données de la vue"""

        if view not in cls.get_views_struct():
            # print(f"structure pour {cls} vue : {view} non defini")
            return None

        if prop not in cls.get_views_struct()[view]:
            if prop == "fields":
                return "__all__"

            if view == "create" and prop == "title":
                return _("Nouveau") + " " + cls._meta.verbose_name.lower()

            # print(f"propriete {prop} non definie pour {cls} vue : {view}")
            return None

        return cls.get_views_struct()[view][prop]

    @classmethod
    def get_list_menu(cls):
        views_struct = cls.get_views_struct()
        assert "list" in views_struct, f"No list view for {cls}"
        assert "menu" in views_struct["list"], f"No menu in listview for {cls}"

        list_menu = []
        for model_ref, view_name, faclass, _ in views_struct["list"]["menu"]:
            model = get_model(model_ref)
            list_menu.append((model.get_url(view_name.lower()), faclass))
        return list_menu

    @classmethod
    def get_class_menu(cls, view):
        views_struct = cls.get_views_struct()
        assert view in views_struct, f"No {view} view for {cls}"
        if "menu" not in views_struct[view]:
            return None

        menu_tuple = views_struct[view]["menu"]

        class_menu = []
        for menu in menu_tuple:
            model = get_model(menu[0])
            # pas de tooltip
            if len(menu) == 4:
                class_menu.append((model.get_url(menu[1].lower()), menu[2], False))
            # tooltip
            else:
                class_menu.append((model.get_url(menu[1].lower()), menu[2], menu[4]))

        if "extra_menu" in views_struct[view]:
            extras = views_struct[view]["extra_menu"]
            for url, faclass, _, _, tooltip in extras:
                # i = 0
                # if url:
                # href = reverse_lazy(url)
                # else:
                # href = "#"
                class_menu.append((url, faclass, tooltip))

        return class_menu

    # def get_detail_menu(self):
    #     assert 'detail' in self.__class__.get_views_struct(), f"No detail view for {self.__class__}"
    #     assert 'menu' in self.__class__.get_views_struct()['detail'], f"No detail menu for {self.__class__}"

    #     detail_menu = []

    #     for model_ref, view_name, faclass, url_args in self.__class__.get_views_struct()['detail']['menu']:

    #         module_name, class_name = model_ref.split(':')
    #         module = importlib.import_module(module_name)
    #         model = getattr(module, class_name)

    #         kwargs = {}
    #         for self_field, other_field in url_args:
    #             if not other_field:
    #                 if not hasattr(model, self_field):
    #                     print(f"Warning : trying to acces non existent url arg {field} in {self.__class__} when generating detail menu")
    #                     continue
    #                 kwargs[self_field] = getattr(self, self_field)
    #             else:
    #                 assert hasattr(self, self_field)
    #                 assert hasattr(model, other_field)
    #                 kwargs['param'] = other_field
    #                 kwargs['pk'] = getattr(self, self_field)

    #         href = reverse_lazy(model.get_url(view_name.lower()), kwargs=kwargs)
    #         detail_menu.append( (href, faclass) )
    #     return detail_menu

    def get_view_menu(self, view):
        assert (
            view in self.__class__.get_views_struct()
        ), f"No {view} view for {self.__class__}"
        if "menu" not in self.__class__.get_views_struct()[view]:
            return None

        struct_menu = self.__class__.get_views_struct()[view]["menu"]
        built_menu = []

        # for model_ref, view_name, faclass, url_args in struct_menu:
        for menu in struct_menu:
            model = get_model(menu[0])

            kwargs = {}
            i = 0
            for self_field, other_field in menu[3]:
                if not other_field:
                    if not hasattr(model, self_field):
                        print(
                            f"Warning : trying to acces non existent url arg {self_field} or {other_field} in {self.__class__} when generating detail menu"
                        )
                        continue
                    kwargs[self_field] = getattr(self, self_field)
                else:
                    assert hasattr(self, self_field)
                    assert hasattr(model, other_field)
                    kwargs["param" + str(i)] = other_field
                    kwargs["pk" + str(i)] = getattr(self, self_field)
                i = i + 1

            href = reverse_lazy(model.get_url(menu[1].lower()), kwargs=kwargs)
            # pas de tooltip
            if len(menu) == 5:
                built_menu.append((href, menu[2], None, None, menu[4]))
            else:
                built_menu.append((href, menu[2], None, None, None))

        if "extra_menu" in self.__class__.get_views_struct()[view]:
            extras = self.__class__.get_views_struct()[view]["extra_menu"]
            for url, url_args, faclass, aclass, data, tooltip in extras:

                kwargs = {}
                i = 0
                for self_field, other_field in url_args:
                    if not other_field:
                        if not hasattr(model, self_field):
                            print(
                                f"Warning : trying to acces non existent url arg {self_field} or {other_field} in {self.__class__} when generating detail menu"
                            )
                            continue
                        kwargs[self_field] = getattr(self, self_field)
                    else:
                        assert hasattr(self, self_field)
                        assert hasattr(model, other_field)
                        kwargs["param" + str(i)] = other_field
                        kwargs["pk" + str(i)] = getattr(self, self_field)
                    i = i + 1

                if url:
                    href = reverse_lazy(url, kwargs=kwargs)
                else:
                    href = "#"
                built_menu.append((href, faclass, aclass, data, tooltip))

        return built_menu

    @classmethod
    def get_url(cls, view_name):
        assert (
            view_name in cls.get_views_struct().keys()
        ), f"trying to reverse url with a view not specified in views_struct for {cls} with view {view_name}"
        return f"{cls._meta.app_label}:{cls.__name__.lower()}-" + view_name

    # @classmethod
    # def get_fields_data(cls, fieldlist):
    #     field_data = []
    #     for field in fieldlist:
    #         if not hasattr(cls, field):
    #             continue
    #         field_data.append( (field, getattr(cls, field).field.verbose_name) )
    #     print(field_data)
    #     return field_data

    @classmethod
    def extract_json_dict(cls, json):
        kwargs = {}
        for key, value in json.items():
            related_model = getattr(cls, key).field.related_model
            if related_model:
                kwargs[key] = related_model.objects.get(pk=value)
            else:
                kwargs[key] = value
        return kwargs

    @classmethod
    def extract_record_dict(cls, record):
        kwargs = {}
        for field in record:
            splitted_field = field.split("__")
            if len(splitted_field) > 1:
                model_name = splitted_field[0]
                # model_field = splitted_field[1]
                related_model = getattr(cls, model_name).field.related_model
                obj, _ = related_model.objects.get_or_create(name=record[field])
                kwargs[model_name] = obj
            else:
                kwargs[field] = record[field]
        print(kwargs)
        return kwargs

    #     @classmethod
    #     def extract_record_dict(cls, record):
    #         kwargs = {}
    #         for field in record:
    #             related_model = getattr(cls, field).field.related_model
    #             if related_model:
    #                 kwargs[field] = related_model.objects.get(pk=record[field])
    #             else:
    #                 kwargs[field] = record[field]
    #         return kwargs

    @classmethod
    def create_object_list_from_records(cls, records):
        object_list = []
        # index = 0
        # print(fields)
        for record in records:
            # print(record)
            o = cls(**cls.extract_record_dict(record))
            o.save()
            # o_dict = {}
            # for field in fields:
            #     related_model = getattr(cls, field).field.related_model
            #     print(related_model)
            #     if related_model:
            #         o_dict[field] = getattr(o, field).pk
            #     else:
            #         o_dict[field] = getattr(o, field)
            # o.json = json.dumps(o_dict)
            # o.index = index
            # index += 1
            object_list.append(o)
        return object_list

    # @classmethod
    # def get_samples(cls):
    #     return [cls(), cls()]
    # @classmethod
    # def get_omnibar_qs(cls):
    #     return cls.objects.all()

    # @classmethod
    # def get_id(cls):
    #     return cls.__name__.lower()


class GenericHistory(GenericModel):
    """
    Recense tous les changements important
    """

    ACTION_CHOICES = [
        ("ED", _("Modification")),
        ("DE", _("Suppression")),
        ("NE", _("Création")),
    ]

    action = models.CharField(max_length=2, choices=ACTION_CHOICES, default="ED")
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, null=True, related_name="global_history"
    )
    object_id = models.PositiveIntegerField(null=True)
    content_object = GenericForeignKey()

    record_name = models.CharField(max_length=255, null=True, blank=True)
    model_verbose_name = models.CharField(max_length=255, null=True, blank=True)

    date = models.DateTimeField(auto_now=True)

    views_struct = {
        "datatable": {
            "title": _("Historique"),
            "menu": [],
            "ordering": ["-date"],
            "fields": [
                ("action", "action", "Action"),
                ("date", "date", "Date"),
                ("user_link", "user.username", "Utilisateur"),
                ("record_link", "record_name", "Objet"),
                ("model_list_link", "model_verbose_name", "Type"),
            ],
            "api_url": "api/globalhistory",
            "css_modal_id": "modal-changes",
        },
        # 'list': {
        #     'tpl': 'django_bkendoz/globalhistory_list.html',
        #     'title': _("Historique"),
        #     'menu': [],
        #     'ordering': ['-date']
        # },
        "detail": {
            "tpl": "django_bkendoz/history_changes.html",
            "menu": [],
        },
    }

    class Meta:
        abstract = True

    @classmethod
    def update(cls, action, user, m2m_changes, content_object):
        hist = cls(
            action=action,
            user=user,
            content_object=content_object,
            record_name=str(content_object.instance),
            model_verbose_name=content_object.instance.__class__._meta.verbose_name,
        )
        hist.save()

        last_change = content_object
        prev_change = content_object.prev_record
        # if not prev_change:
        #     return []
        for change in last_change.diff_against(prev_change).changes:
            f = content_object.instance.__class__._meta.get_field(change.field)
            if f.is_relation:
                if change.old:
                    old = f.related_model.objects.get(pk=change.old)
                else:
                    old = None
                new = f.related_model.objects.get(pk=change.new)
            else:
                old = change.old
                new = change.new

            hist_change = get_history_change_model()(
                field_name=change.field, old_value=old, new_value=new, history=hist
            )
            hist_change.save()

        for field, values in m2m_changes.items():
            old = str_list_to_str([str(v) for v in values["old"]])
            new = str_list_to_str([str(v) for v in values["new"]])
            hist_change = get_history_change_model()(
                field_name=field, old_value=old, new_value=new, history=hist
            )
            hist_change.save()

        return hist

    def get_history_changes(self):
        changes = get_history_change_model().objects.filter(history=self)
        return changes


class GenericHistoryChange(models.Model):
    """
    Stock les changements lors des mises à jour
    Ajouter l'attribut foreign key history dans la class héritée
    """

    field_name = models.CharField(max_length=255)
    old_value = models.TextField()
    new_value = models.TextField()

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        if not self.old_value:
            self.old_value = ""
        if not self.new_value:
            self.new_value = ""
        super().save(args, kwargs)


class GenericHistorizedModel(GenericModel):
    # history_date = None
    # history_object
    history = HistoricalRecords(inherit=True)

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        # self.history_date = None
        # self.history = None
        super().__init__(*args, **kwargs)

    @property
    def hist_date(self):
        return self.history_date

    def history_create(self, user, *args, **kwargs):
        content_object = self.history.first()
        hist = get_global_history_model()(
            action="NE",
            user=user,
            content_object=content_object,
            record_name=str(content_object.instance),
            model_verbose_name=content_object.instance.__class__._meta.verbose_name,
        )
        hist.save()

    def history_update(self, user, m2m_changes=None):
        content_object = self.history.first()
        get_global_history_model().update(
            "ED",
            user,
            m2m_changes,
            content_object,
        )
        # hist.save()

    def history_delete(self, user, *args, **kwargs):
        content_object = self.history.first()
        hist = get_global_history_model()(
            action="DE",
            user=user,
            content_object=content_object,
            record_name=str(content_object.instance),
            model_verbose_name=content_object.instance.__class__._meta.verbose_name,
        )
        hist.save()


class GenericExcelModel:
    pass



import json

import pyexcel
from django.conf import settings
from django.contrib.auth import get_user_model, views as auth_views
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.db import models
from django.http import FileResponse, JsonResponse
from django.shortcuts import redirect
from django.urls import path, reverse_lazy
from django.views.generic import (CreateView, DeleteView, DetailView, ListView,
                                  UpdateView)
from django.views.generic.base import TemplateView, View
from django.views.generic.edit import FormView
from rest_framework import status, viewsets
from rest_framework.permissions import IsAdminUser
from rest_framework.response import Response
from rest_framework.views import APIView

from .core import (get_color_from_class, get_global_history_model, get_href,
                   get_layout_data, get_model_fields)
from .forms import ImportXlsForm, get_generic_form
from .models import GenericModel
from .serializers import GenericSerializer, GenericUserSerializer


class GenericDashboardView(PermissionRequiredMixin, TemplateView):
    """
    Vue en tuile prenant en entrée une liste de modèles.
    """

    # models: list[GenericModel] = []
    models = []
    template_name = "django_bkendoz/dashboard.html"
    dflt_model_template = "django_bkendoz/dashboard_tpl.html"
    dflt_tile_class = "tile-simple-column"
    list_cls = "dashboard-list"
    auth = "superuser"

    def __init__(self, **kwargs):
        assert hasattr(
            settings, "DASH_LIST_COUNT"
        ), """
            DASH_LIST_COUNT non défini dans settings.py,
            il faut définir la valeur par default du nombre d'element à afficher dans le tableau de bord
            """
        self.permission_required = f"{self.auth}.dashboard"
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["models"] = {}
        for m in self.models:
            m_dict = context["models"][m] = {}
            o_cnt = m.get_view_data("dashboard", "count") or settings.DASH_LIST_COUNT
            o_ord = m.get_view_data("dashboard", "orderby") or None
            m_dict["title"] = m.get_view_data("dashboard", "title")
            m_dict["class"] = (
                m.get_view_data("dashboard", "class") or self.dflt_tile_class
            )
            m_dict["html"] = (
                m.get_view_data("dashboard", "html") or self.dflt_model_template
            )
            m_dict["menu"] = m.get_class_menu("dashboard")
            m_dict["fields"] = self.get_dash_fields(m)
            if o_ord:
                m_dict["list"] = m.objects.all().order_by(o_ord)[:o_cnt]
            else:
                m_dict["list"] = m.objects.all()[:o_cnt]

        context["layout_data"] = get_layout_data()
        context["settings"] = settings
        return context

    def get_dash_fields(self, model):
        """
        Retourne les champs à afficher dans la vue en tuile.
        :param model: model dont les champs sont à afficher
        :type model: GenericModel
        [str]: Liste de champs
        """
        view_struct_fields = model.get_view_data("dashboard")
        return get_model_fields(model, view_struct_fields)


class GenericOmnibarView(View):
    models = []
    data = []
    auth = "superuser"

    def __init__(self, **kwargs):
        self.permission_required = f"{self.auth}.omnibar"
        super().__init__(**kwargs)

    def get(self, request, *args, **kwargs):
        items = {}
        for m in self.models:
            assert issubclass(
                m, GenericModel
            ), f"""
                Pour utiliser la vue Omnibar la classe {m.get_id()} doit hériter de GenericModel'
            """
            key = m._meta.label
            items[key] = []
            cached = cache.get("omnibar:" + key)
            if not cached:
                cached = m.to_omnibar_dict()
                cache.set("omnibar:" + key, cached)
            items[key] = cached
        return JsonResponse(items, safe=False)


class GenericViewsetView(viewsets.ModelViewSet):
    queryset = None
    serializer_class = None
    model = None

    def get_queryset(self):
        return self.model.objects.all().order_by('-id')

    def get_serializer_class(self):
        if not self.serializer_class:
            serializer_class = GenericSerializer
            serializer_class.Meta.model = self.model
            self.serializer_class = serializer_class
        return self.serializer_class

    def list(self, request, format=None):
        # pas de cache si recherche
        qp = request.query_params
        if "search[value]" in qp and qp["search[value]"]:
            return super().list(request, format)

        if "start" not in qp or "length" not in qp:
            return super().list(request, format)

        key = f"{qp['start']}:{qp['length']}"
        cached = cache.get(self.model.__name__ + ":viewset:" + key)
        if not cached:
            cached = super().list(request, format).data
            cache.set(self.model.__name__ + ":viewset:" + key, cached)
        return Response(cached)

    class Meta:
        abstract = True


class GenericDatatableView(PermissionRequiredMixin, ListView):
    model: GenericModel
    template_name = "django_bkendoz/datatable_list.html"
    css_class = "datatable"
    css_modal_id = ""
    api_url: str

    def __init__(self, **kwargs):
        model_name = self.model.__name__

        #     assert globals().get(viewset), f'''
        # Erreur : impossible de créer la vue datatable pour le modèle {model_name} car aucun viewset {viewset} n'existe.
        #     '''

        self.current_app = self.model._meta.app_label
        self.permission_required = f"{self.current_app}.datatable_{model_name.lower()}"
        self.template_name = (
            self.model.get_view_data("datatable", "tpl") or self.template_name
        )

        reverse_url = f"{self.current_app}:{self.model.__name__.lower()}-viewset"
        self.api_url = self.model.get_view_data("datatable", "api_url") or reverse_lazy(
            reverse_url
        )
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context["title"] = self.model.get_view_data("datatable", "title")
        context["menu"] = self.model.get_class_menu("datatable")
        context["css_class"] = self.model.get_view_data("datatable", "css_class") or self.css_class
        context["css_modal_id"] = self.model.get_view_data("datatable", "css_modal_id") or self.css_modal_id
        context["fields"] = self.model.get_view_data("datatable", "fields")
        context["api_url"] = self.api_url

        context["settings"] = settings
        context["layout_data"] = get_layout_data()
        return context


class GenericListView(PermissionRequiredMixin, ListView):
    template_name = "django_bkendoz/list.html"
    template_list_name = "django_bkendoz/table.html"
    list_cls = "generic-list"
    fields = "__all__"
    ordering = ["pk"]

    def __init__(self, **kwargs):
        assert hasattr(
            settings, "LIST_FIELDMAP"
        ), f"""
            Pas de template LIST_FIELDMAP défini dans settings.py,
            il faut définir comment chaque field de la liste va s'afficher dans un fichier de template .html
            """
        self.current_app = self.model._meta.app_label
        self.permission_required = (
            f"{self.current_app}.list_{self.model.__name__.lower()}"
        )
        self.ordering = self.model.get_view_data("list", "ordering") or self.ordering
        self.template_name = (
            self.model.get_view_data("list", "tpl") or self.template_name
        )
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        context["list_title"] = self.model.get_view_data("list", "title")
        context["hide_search"] = (
            self.model.get_view_data("list", "hide_search") or False
        )
        context["list_menu"] = self.model.get_class_menu("list")
        context["list_cls"] = self.list_cls
        context["fields"] = self._get_fields()
        context["datatable_fields"] = self.model.get_view_data("viewset", "fields")
        context["api_url"] = f"api/{self.model.__name__.lower()}s/"
        context["html"] = (
            self.model.get_view_data("list", "html") or self.template_list_name
        )

        context["settings"] = settings
        context["layout_data"] = get_layout_data()
        return context

    def _get_fields(self):
        """Champs à afficher dans la vue en liste.

        Returns:
            str: Liste des champs
        """
        fields = self.model.get_view_data("list", "fields")
        return get_model_fields(self.model, fields)


class GenericCreateView(PermissionRequiredMixin, CreateView):
    template_name = "django_bkendoz/form_view.html"
    color_classes = []

    def __init__(self, **kwargs):
        assert hasattr(
            settings, "FORM_FIELDMAP"
        ), f"""
            Pas de template FORM_FIELDMAP défini dans settings.py,
            il faut définir dans ce fichier html comment chaque field
            particulier des formulaires vont s'afficher.
            """
        self.current_app = self.model._meta.app_label
        self.permission_required = (
            f"{self.current_app}.add_{self.model.__name__.lower()}"
        )
        self.template_name = (
            self.model.get_view_data("create", "html") or self.template_name
        )

        if not self.form_class:
            self.form_class = self._create_generic_form()
            self.initial = {"model": self.model}
            # both fields and form_class not permitted
            self.fields = None
            # indent ? TODO
            super().__init__(**kwargs)

    def _create_generic_form(self):
        view_struct_fields = self.model.get_view_data("create")
        return get_generic_form(self.model, view_struct_fields)

    def form_valid(self, form):
        if hasattr(self.model, "history"):
            new_object = form.save()
            self.object = new_object
            new_object.history_create(
                self.request.user,
            )
            return redirect(self.get_success_url())
        return super().form_valid(form)

    def get_context_data(self, **kwargs):
        if "param0" in self.kwargs and "pk0" in self.kwargs:
            self.initial = {self.kwargs["param0"]: self.kwargs["pk0"]}
        if "param1" in self.kwargs and "pk1" in self.kwargs:
            self.initial = {
                self.kwargs["param0"]: self.kwargs["pk0"],
                self.kwargs["param1"]: self.kwargs["pk1"],
            }

        context = super().get_context_data(**kwargs)

        context["menu"] = self.model.get_class_menu("create")
        context["form_type"] = "create"
        context["form_title"] = self.model.get_view_data("create", "title")
        context["form_action"] = reverse_lazy(
            f"{self.current_app}:{self.model.__name__.lower()}-create"
        )
        if len(self.color_classes) > 0:
            context["color_palette"] = get_color_from_class(self.color_classes)

        context["layout_data"] = get_layout_data()
        context["settings"] = settings
        return context

    def get_success_url(self):
        self.absolute_success_url = self.model.get_view_data(
            "create", "absolute_success_url"
        )
        if self.absolute_success_url:
            return reverse_lazy(self.absolute_success_url)

        self.success_url = (
            get_href(self.model.get_view_data("create", "success_url"), self.object)
            or self.success_url
        )
        if self.success_url:
            return self.success_url

        for v, args in (("detail", [self.object.pk]), ("list", None), ("create", None)):
            if self.model.get_view_data(v):
                url = f"{self.current_app}:{self.model.__name__.lower()}-{v}"
                return reverse_lazy(url, args=args)


class GenericUpdateView(PermissionRequiredMixin, UpdateView):
    template_name = "django_bkendoz/form_view.html"
    color_classes = []

    def __init__(self, **kwargs):
        assert hasattr(
            settings, "FORM_FIELDMAP"
        ), f"""
            Pas de template FORM_FIELDMAP défini dans settings.py,
            il faut définir dans ce fichier html comment chaque field
            particulier des formulaires vont s'afficher.
            """
        self.current_app = self.model._meta.app_label
        self.permission_required = (
            f"{self.current_app}.change_{self.model.__name__.lower()}"
        )
        self.template_name = (
            self.model.get_view_data("update", "html") or self.template_name
        )
        if not self.form_class:
            self.form_class = self._create_generic_form()
            # fields et form_class ne peuvent etre attribues en meme temps
            self.fields = None
            self.initial = {"model": self.model}
            super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["form_type"] = "update"
        context["form_title"] = self.model.get_view_data("update", "title")
        context["form_action"] = reverse_lazy(
            f"{self.current_app}:{self.model.__name__.lower()}-update",
            args=[self.object.pk],
        )
        if len(self.color_classes) > 0:
            context["color_palette"] = get_color_from_class(self.color_classes)

        context["menu"] = self.object.get_view_menu("update")
        context["settings"] = settings
        context["layout_data"] = get_layout_data()
        return context

    def form_valid(self, form):
        if hasattr(self.model, "history"):
            m2m_changes = self._compare_m2m_changes(form)
            self.object = form.save()
            self.object.history_update(self.request.user, m2m_changes)
            # form.save_m2m()
            return redirect(self.get_success_url())
        return super().form_valid(form)

    def get_success_url(self):
        self.success_url = (
            get_href(self.model.get_view_data("update", "success_url"), self.object)
            or self.success_url
        )
        if self.success_url:
            return self.success_url

        for v, args in (("detail", [self.object.pk]), ("list", None), ("create", None)):
            if self.model.get_view_data(v):
                url = f"{self.current_app}:{self.model.__name__.lower()}-{v}"
                return reverse_lazy(url, args=args)

    def _create_generic_form(self):
        view_struct_fields = self.model.get_view_data("update")
        return get_generic_form(self.model, view_struct_fields)

    def _compare_m2m_changes(self, form):
        m2m_changes = {}
        for f in self.model._meta.get_fields():
            if f.many_to_many and not isinstance(f, models.ManyToManyRel):
                m2m_changes[f.name] = {
                    "old": set(getattr(self.object, f.name).all()),
                    "new": set(form.cleaned_data[f.name]),
                }
                if m2m_changes[f.name]["old"] == m2m_changes[f.name]["new"]:
                    m2m_changes.pop(f.name)
        return m2m_changes


class GenericDetailView(PermissionRequiredMixin, DetailView):
    template_name = "django_bkendoz/detail.html"

    def __init__(self, **kwargs):
        assert hasattr(
            settings, "DETAIL_FIELDMAP"
        ), f"Pas de template DETAIL_FIELDMAP défini dans settings.py, il faut définir comment chaque field de la vue de détail va s'afficher dans un template"
        self.current_app = self.model._meta.app_label
        self.permission_required = (
            f"{self.current_app}.detail_{self.model.__name__.lower()}"
        )
        self.template_name = (
            self.model.get_view_data("detail", "tpl") or self.template_name
        )
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["detail_menu"] = self.object.get_view_menu("detail")
        context["fields"] = self._get_fields()
        context["html"] = self.model.get_view_data("detail", "html") or None
        context["settings"] = settings
        context["MEDIA_URL"] = settings.MEDIA_URL
        context["layout_data"] = get_layout_data()
        return context

    def _get_fields(self):
        fields = self.model.get_view_data("detail", "fields")
        if fields == "__all__":
            return None
        return fields


class GenericHistoryView(PermissionRequiredMixin, DetailView):
    template_name = "django_bkendoz/history.html"

    def __init__(self, **kwargs):
        self.current_app = self.model._meta.app_label
        self.permission_required = (
            f"{self.current_app}.history_{self.model.__name__.lower()}"
        )
        self.template_name = (
            self.model.get_view_data("history", "tpl") or self.template_name
        )
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["list_title"] = self.model.get_view_data("history", "title")
        context["history_menu"] = self.object.get_view_menu("history")
        context["fields"] = self._get_fields()
        context["html"] = self.model.get_view_data("history", "html") or None

        ct = ContentType.objects.get_for_model(self.object)
        context["history"] = get_global_history_model().objects.filter(
            record_model=ct,
            record_id=self.object.pk,
        )

        context["settings"] = settings
        context["layout_data"] = get_layout_data()
        return context

    def _get_fields(self):
        fields = self.model.get_view_data("history", "fields")
        if fields == "__all__":
            return None
        return fields


class GenericDeleteView(PermissionRequiredMixin, DeleteView):
    template_name = "django_bkendoz/delete.html"

    def __init__(self, **kwargs):
        self.current_app = self.model._meta.app_label
        self.permission_required = (
            f"{self.current_app}.delete_{self.model.__name__.lower()}"
        )
        self.template_name = (
            self.model.get_view_data("delete", "tpl") or self.template_name
        )
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["back_url"] = f"{self.current_app}:{self.model.__name__.lower()}-detail"
        context["html"] = self.model.get_view_data("delete", "html") or None
        context["settings"] = settings
        context["layout_data"] = get_layout_data()
        return context

    def delete(self, *args, **kwargs):
        if hasattr(self.model, "history"):
            self.object = self.get_object()
            self.object.history_delete(self.request.user)
        return super().delete(*args, **kwargs)

    def get_success_url(self):
        for v, args in (("list", None), ("create", None)):
            if self.model.get_view_data(v):
                url = f"{self.current_app}:{self.model.__name__.lower()}-{v}"
                return reverse_lazy(url, args=args)


class GenericExportxlsView(PermissionRequiredMixin, View):
    model = None

    def __init__(self, **kwargs):
        self.current_app = self.model._meta.app_label
        self.permission_required = (
            f"{self.current_app}.exportxls_{self.model.__name__.lower()}"
        )
        super().__init__(**kwargs)

    def get(self, request, *args, **kwargs):
        if hasattr(self.model, "get_excel_dict"):
            xls_file = pyexcel.save_as(
                dest_file_type="xls", adict=self.model.get_excel_dict()
            )
            return FileResponse(
                xls_file,
                as_attachment=True,
                filename=f"{self.model.__name__.lower()}_export.xls",
            )
        else:
            tuple_fields = self.get_exportxls_fields()
            tuple_fields.reverse()
            if tuple_fields == "__all__":
                objects = self.model.objects.values()
            else:
                objects = self.model.objects.values(*tuple_fields)

            xls_file = pyexcel.save_as(dest_file_type="xls", records=objects)
            return FileResponse(
                xls_file,
                as_attachment=True,
                filename=f"{self.model.__name__.lower()}_export.xls",
            )

    def get_exportxls_fields(self):
        view_struct_fields = self.model.get_view_data("exportxls")
        return view_struct_fields


class GenericImportxlsView(PermissionRequiredMixin, FormView):
    template_name = "django_bkendoz/import_xls.html"
    form_class = ImportXlsForm
    model = None
    fields = []

    def __init__(self, **kwargs):
        self.current_app = self.model._meta.app_label
        self.permission_required = (
            f"{self.current_app}.importxls_{self.model.__name__.lower()}"
        )
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["form_title"] = f"Importer des données excel"
        context["form_subtitle"] = self.model._meta.verbose_name
        context["fields"] = self.get_import_fields()
        context["import_sample"] = self.model.get_samples()
        context["layout_data"] = get_layout_data()
        context["settings"] = settings
        return context

    def get_import_fields(self):
        view_struct_fields = self.model.get_view_data("importxls")
        return get_model_fields(self.model, view_struct_fields)

    def form_valid(self, form):
        xls_file = form.files["xls_file"]
        # print(pyexcel.get_sheet(file_type='xls', file_content=xls_file))
        records = pyexcel.get_records(file_type="xls", file_content=xls_file)

        context = self.get_context_data()
        # context['preview'] = True
        context["object_list"] = self.model.create_object_list_from_records(
            records, self.fields
        )
        # context['fields'] = get_model_fields(self.model, self.fields)
        # context['model_plural'] = self.model._meta.verbose_name_plural
        # context['form_action'] = f"{self.current_app}:{self.model.__name__.lower()}-importxlssave"
        return super().form_valid(form)
        # return render(self.request, self.template_name, context=context)

    def get_success_url(self):
        self.absolute_success_url = self.model.get_view_data(
            "importxls", "absolute_success_url"
        )
        if self.absolute_success_url:
            return reverse_lazy(self.absolute_success_url)

        for v, args in (("list", None), ("create", None)):
            if self.model.get_view_data(v):
                url = f"{self.current_app}:{self.model.__name__.lower()}-{v}"
                return reverse_lazy(url, args=args)


class GenericImportxlssaveView(PermissionRequiredMixin, View):
    model: GenericModel

    def post(self, request):
        for key, value in request.POST.items():
            if key.startswith("chk_"):
                if value == "off":
                    continue

                index = key.split("_")[1]
                data = json.loads(request.POST[index])
                m = self.model(**self.model.extract_json_dict(data))
                m.save()

        return redirect(
            reverse_lazy(f"{self.current_app}:{self.model.__name__.lower()}-list")
        )

    def __init__(self, **kwargs):
        self.current_app = self.model._meta.app_label
        self.permission_required = (
            f"{self.current_app}.importxlssave_{self.model.__name__.lower()}"
        )
        super().__init__(**kwargs)


#   def form_valid(self, form):
#      def testinit(row):
#         # self.model.book_category.field.related_model
#         i = 0
#         new_row = []
#         for field_name in self.fields:
#            related_model = getattr(self.model, field_name).field.related_model
#            if related_model:
#               new_row.append(related_model.objects.get(pk=row[i]))
#            else:
#               new_row.append(row[i])
#            i = i + 1

#         print(new_row)
#         return new_row

#      xls_file = form.files['xls_file']
#      #print(pyexcel.get_sheet(file_type='xls', file_content=xls_file))
#      pyexcel.save_as(file_content=xls_file, file_type='xls',
#                            dest_initializer=testinit,
#                            dest_model=self.model,
#                       start_row=1,
#                       colnames=["book_category", "name"],
#                       dest_mapdict=["book_category", "name"]
#                      )

#      return super().form_valid(form)

#    def get_success_url(self):
#        if self.success_url:
#           return reverse_lazy(self.success_url)
#        else:
#           current_app = self.request.resolver_match.namespace
#           return reverse_lazy(f"{current_app}:{self.model.__name__.lower()}-importxls-preview")

#   def get(self, request, *args, **kwargs):
#      if self.fields == '__all__':
#         objects = self.model.objects.values()
#      else:
#         objects = self.model.objects.values(*self.fields)

#      xls_file = pyexcel.save_as(dest_file_type="xls", records=objects)
#      response = FileResponse(xls_file, as_attachment=True,
#                              filename=f"{self.model.__name__.lower()}_import.xls")
#      return response
#      # return HttpResponse(xls_file, content_type="application/vnd.ms-excel")
#   pass


class GenericLoginView(auth_views.LoginView):
    extra_context = {"settings": settings, "test": "test"}


class GenericLogoutView(auth_views.LogoutView):
    extra_context = {"settings": settings, "test": "test"}


class GenericProfileView(PermissionRequiredMixin, UpdateView):
    template_name = "auth/user_detail.html"
    model = get_user_model()
    # fields = ["first_name", "last_name", "email", "style"]
    fields = ["first_name", "last_name", "email"]
    extra_context = {
        "settings": settings,
    }

    def __init__(self, **kwargs):
        self.current_app = self.model._meta.app_label
        self.permission_required = (
            f"{self.current_app}.detail_{self.model.__name__.lower()}"
        )
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["settings"] = settings
        context["layout_data"] = get_layout_data()
        context["history"] = get_global_history_model().objects.filter(user=self.object)
        return context


class UserRecordView(APIView):
    """
    API View to create or get a list of all the registered
    users. GET request returns the registered users whereas
    a POST request allows to create a new user.
    """

    permission_classes = [IsAdminUser]

    def get(self, format=None):
        users = get_user_model().objects.all()
        serializer = GenericUserSerializer(users, many=True)
        return Response(serializer.data)

    def post(self, request):
        serializer = GenericUserSerializer(data=request.data)
        if serializer.is_valid(raise_exception=ValueError):
            serializer.create(validated_data=request.data)
            return Response(serializer.data, status=status.HTTP_201_CREATED)

        return Response(
            {
                "error": True,
                "error_msg": serializer.error_messages,
            },
            status=status.HTTP_400_BAD_REQUEST,
        )


def create_model_paths(model):
    mod_name = model.__name__
    mod_name_low = mod_name.lower()
    path_to_return = []

    def get_fields(view):
        fields = model.get_view_data(view.lower())
        if fields == "__all__":
            return get_model_fields(model)
        else:
            return fields
        if view == "Exportxls":
            return get_model_fields(model, fields)

    assert issubclass(
        model, GenericModel
    ), f"""
        {model} doit hériter de GenericModel pour pouvoir créer la structure de vues"""

    for view in model.get_views_struct().keys():
        view = view.title()
        assert globals().get(
            f"Generic" + view + "View"
        ), f"""
    Erreur : aucune vue {view} définit dans la structure de vues pour Generic{view}View pour le modèle {model}.
        """

        ClassView = type(
            # name
            mod_name + view + "View",
            # subclasses
            (globals()["Generic" + view + "View"],),
            # attr
            {"model": model, "fields": get_fields(view)},
        )

        if view == "Viewset":
            path_to_return.append(
                path(
                    "datatables/" + mod_name_low + "s",
                    ClassView.as_view({"get": "list"}),
                    name=mod_name_low + "-" + view.lower(),
                )
            )

        else:

            view_path = mod_name_low + "s/" + view.lower()
            if view in ["Update", "Detail", "Delete", "History"]:
                view_path += "/<int:pk>"

            path_to_return.append(
                path(
                    view_path,
                    ClassView.as_view(),
                    name=mod_name_low + "-" + view.lower(),
                )
            )

            if view == "Create":
                view_path += "/<str:param0>/<int:pk0>"
                view_path1 = view_path + "/<str:param1>/<int:pk1>"
                path_to_return.append(
                    path(
                        view_path,
                        ClassView.as_view(),
                        name=mod_name_low + "-" + view.lower(),
                    )
                )

                path_to_return.append(
                    path(
                        view_path1,
                        ClassView.as_view(),
                        name=mod_name_low + "-" + view.lower(),
                    )
                )

    return path_to_return
